import requests
import json
import datetime
from functools import reduce
from erp.assets.variables import Variables
from erp.assets.networkfunctions import get_client_ip
from erp.models import Loja

#Nao sei se vou precisar ou a class vai funcionar puramente na hora da instancia.
from erp.models import Pedido, Profile, Pagamento

#Modulo para gerenciar pagamentos via Asaas
#A principio o webhook sera gerenciado fora daqui. Mas nao e definitivo, pois vou
#precisar de URls e views para lidar com o retorno e mudar o estado do pagamento e dos
#pedidos. de repente agente faz tudo aqui para ser mais simples.

class Asaas():
    #Requires a payment model -> still figuring it out if we are better keeping it in the main model, or create a separated file for reuse
    
    #Simple Function to allow dig down into tiered json returned data and avoid error in case of missing keys
    def __init__(self, apiToken='', webHookToken = '', production = False):
        self.vars = Variables()
        #Prepara o header e gera as base do URL para usar nos metodos
        self.apiToken = apiToken
        self.webHookToken = webHookToken
        if production:
            self.baseURL = self.vars.ASAAS_PRODUCTION_BASE_URL
        else:
            self.baseURL = self.vars.ASAAS_SANDBOX_BASE_URL
        
        self.headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "access-token" : self.apiToken, 
        }
        self.jsonHeaders = json.dumps(self.headers) #this is the call ready header
    
    #Not sure if we will have a single one, or one for card, pix and boleto.
    #May end up with individual methods for now I am trying to do a single method.
    def nova_cobranca(self, request, pedido): #Pedido e capaz de ser redundante. tenho de ver no uso pratico da class
        """
        Mandatory fields:
            customer
            billingType
            value
            dueDate
        Outras que vamos usar:
            description
            externalReference
            callback (em caso precisamos escrever uma pagina para dizer q foi tudo bem) - E um objeto
                successURL -> not sure we will use this.
        """
        #This assumes asaasid has been properly created in the view. To avoid self calling and having too many entries in the Asaas site for a client
        customer = pedido.cliente.profile.asaasid #Maybe request.user.profile.asaasif seja melhor.
        billingType = pedido.formapagamento
        value = pedido.valortotal() # need () here to ,ake sure I get a float and not a function to go through JSON
        dueDate = pedido.dataenvio
        externalReference = pedido.pk
        loja_data = Loja.objects.first()
        description = f'Pedido # {pedido.pk} - {loja_data.nome} - Total: R${value} - Entrega para: {dueDate}'
        
        #falta montar o call back
     
        data = {
            'customer': customer,
            'billingType': billingType,
            'value' : value, 
            'dueDate'  : dueDate,
            'externalReference': externalReference,
            'description': description,
        }
     
        jsonValues = json.dumps(data)
        
        #Assemble URL for customer creations:
        url = self.baseURL + '/api/v3/payments'

        #Actually send the request -> POST
        formdata = requests.post(url, data=jsonValues, headers= self.headers)
        statusCode = formdata.status_code
        jsonFormData = formdata.json()

        #Process response. 
        if statusCode == 401:
            return 'error(unauthorized)'
        
        if statusCode == 400: 
            error = jsonFormData.get('errors')
            return 'error('+str(error[0])+')'
               
        if statusCode == 200:
            #Add pagamento ao pagamento ao DB
            pagamento_novo = Pagamento()
            pagamento_novo.pedido = pedido
            pagamento_novo.cliente = request.user
            pagamento_novo.paymentID = jsonFormData.get('id')
            pagamento_novo.status = jsonFormData.get('status')
            pagamento_novo.dtdue = dueDate
            pagamento_novo.valorrecebido = jsonFormData.get('netValue')
            pagamento_novo.valorcobranca = jsonFormData.get('value')
            pagamento_novo.billingtype = jsonFormData.get('billingType')
            pagamento_novo.remoteip = get_client_ip(request)
            pagamento_novo.transactionreceipturl = jsonFormData.get('transactionReceiptUrl')
            pagamento_novo.invoiceurl = jsonFormData.get('invoiceUrl')
            #Acho q nao e necessario daro o invoiceurl levar para as transacoes de todos os tipos
            """
            if billingType == 'BOLETO':
                pagamento_novo.boletourl = jsonFormData.get('bankSlipUrl')
            if billingType == 'PIX':
                pagamento_novo.pixtransaction = jsonFormData.get('pixTransaction')
            """
            pagamento_novo.save()
            pedido.status = 'Cobrando'
            pedido.save()
            
            #Here we return o pagamento para usarmos na view para gerar o html bonito.
            return pagamento_novo
        
        return 'error(unknown_error)'        

        
    #this will not inject a UUID to the asaasid field in the pedido after creating the user in asaas.
    #Basically for all payments test if there is a non-0 string on asaasID, if do, proceed with that ID
    #otherwise, run this method, and proceed with the new ID.
    def create_client(self, user):
        #check if client exist - Basically different than '0'
        if user.profile.asaasid != '0':
            return user.profile.asaasid #Cliente alredy has id in asaas, simply overload it if called anyway
        
        #Assemble data for user creation
        name = user.profile.nome
        phone = user.profile.phone
        cpfCnpj = user.profile.cpf
        email = user.email
        address = user.profile.endereco
        complement = user.profile.complemento
        province = user.profile.bairro
        postalcode = user.profile.cep
        externalReference = user.username
        notificationDisabled = True
            
        data = {
            'name': name,
            'phone': phone,
            'cpfCnpj': cpfCnpj,
            'email' : email, 
            'address'  : address,
            'complement' : complement,
            'province' : province,
            'postalcode' : postalcode,
            'externalreference' : externalReference,
            'notificationDisabled' : notificationDisabled,
        }
        jsonValues = json.dumps(data)
        
        #Assemble URL for customer creations:
        url = self.baseURL + '/api/v3/customers'
        
        #Actually send the request -> POST
        formdata = requests.post(url, data=jsonValues, headers= self.headers)
        statusCode = formdata.status_code
        jsonFormData = formdata.json()
        
        #Process response. 
        if statusCode == 401:
            return 'error(unauthorized)'
        
        if statusCode == 400: 
            error = jsonFormData.get('errors')
            return 'error('+str(error[0])+')'

        
        #usuario criado e aqui tem de escrever o codigo
        if statusCode == 200:
            #return new ID. Processing request will add ID to the DB
            #I do not want the code to use that type of information
            return jsonFormData.get('id', '0')
        
        return 'error(unknown_error)'


    #These methods will get the info straight from Asaas. Not from the DB
    def get_client_id(self, pagamento): #Done
        return pagamento.cliente.profile.asaasid

    def get_pix_code(self, pagamento): # Done
        url = self.baseURL+'api/v3/payments/'+ str(pagamento.paymentid) + '/pixQrCode'
        #Actually send the request -> GET
        formdata = requests.get(url, headers= self.headers)
        statusCode = formdata.status_code
        jsonFormData = formdata.json()
        
        if statusCode == 401:
            return 'error(unauthorized)'
        if statusCode == 403:
            return 'error(invalidbody)'
        if statusCode == 404:
            return 'error(notfound)'
        if statusCode == 200:
            return jsonFormData.get('paylaod') #Copia e Cola pix
        return 'error(unknown)'

    def get_boleto_barcode(self, pagamento): #Done
        url = self.baseURL+'api/v3/payments/'+ str(pagamento.paymentid) + '/identificationField'
        #Actually send the request -> GET
        formdata = requests.get(url, headers= self.headers)
        statusCode = formdata.status_code
        jsonFormData = formdata.json()
        if statusCode == 401:
            return 'error(unauthorized)'
        if statusCode == 403:
            return 'error(invalidbody)'
        if statusCode == 404:
            return 'error(notfound)'
        if statusCode == 200:
            return jsonFormData.get('identificationField')
        return 'error(unknown)'

    def get_status(self, pagamento): #Done
        url = self.baseURL+'api/v3/payments/'+ str(pagamento.paymentid) + '/status'
        #Actually send the request -> GET
        formdata = requests.get(url, headers= self.headers)
        statusCode = formdata.status_code
        jsonFormData = formdata.json()
        
        if statusCode == 401:
            return 'error(unauthorized)'
        if statusCode == 403:
            return 'error(invalidbody)'
        if statusCode == 404:
            return 'error(notfound)'
        if statusCode == 200:
            return jsonFormData.get('status') 
        return 'error(unknown)'
        
              

    def handle_webhook(self, jsondata, token):
        """
        this will have a URL and view to handle te response from asaas when event 
        happens at the bank. Need to setup a token to garantee that all is honest and good to go
        and avoid Fraud. Will do it later.
        
        It is possible to be customized by pagamento. But I plan to keep it simple.
        """
        #Make sure the message matches the token to be sure it is coming from Asaas  
        if self.webHookToken != token:
            return 'error(invalidtoken)'
        
        #Put data together before takin actions
        #jsondata = payload
        event_ID = jsondata.get('id', None)
        status = jsondata.get('event', None)
        payment_ID_obj = jsondata.get('payment', None)
        payment_ID = payment_ID_obj.get('id', None)
        try:
            pagamento = Pagamento.objects.get(paymentID=payment_ID)
        except Pagamento.DoesNotExist:
            pagamento = None
            return 'OK(No Action)'
        
        if event_ID not in pagamento.retornobancario:
            #Se nao achou achou um pagamento con a payment_ID)
            if not pagamento:
                return 'error(payment not found)'
            
            #Atualiza de PENDING ara PAYMENT CREATED once Asaas confirm it.
            if status in ['PAYMENT_CREATED', 'PAYMENT_BANK_SLIP_VIEWED', 'PAYMENT_CHECKOUT_VIEWED']:
                pagamento.retornobancario += datetime.datetime.now().strftime("%Y-%m-%d, %H:%M:%S") + '-'+ event_ID +'-' + status +'\n'
                pagamento.status = status
                pagamento.save()
                return 'OK(Pagamento atualizado)'
                   
            #Dependento do Status do eventprocessar diferente:
            if status in ['PAYMENT_CONFIRMED', 'PAYMENT_RECEIVED', 'PAYMENT_AUTHORIZED']:
                #Do update the payment and pedido
                pagamento.retornobancario += datetime.datetime.now().strftime("%Y-%m-%d, %H:%M:%S") + '-'+ event_ID +'-' + status +'\n'
                pagamento.status = status
                pagamento.controlpago = True
                pagamento.dtpago = datetime.datetime.now()
                if pagamento.pedido.controlpago == False:
                    pedido = pagamento.pedido
                    pedido.status = 'Pago'
                    pedido.dtpago = datetime.datetime.now()
                    pedido.controlpago = True
                    pedido.save()
                pagamento.save()
                return 'OK(Pedido confirmado e pago)'
            
            if status in ['PAYMENT_DELETED', 'PAYMENT_OVERDUE']:
                #Nesses estados temos de marcar o pagamento como ferrado e atualizar o pedido para cancelado                pagamento.retornobancario += event_ID +'\n'
                pagamento.retornobancario += datetime.datetime.now().strftime("%Y-%m-%d, %H:%M:%S") + '-' + event_ID +'-' + status +'\n'
                pagamento.status = status
                pedido = pagamento.pedido
                pedido.status = 'Cancelado'
                pedido.save()
                pagamento.save()
                return 'OK(Pedido e pagamentos cancelados)'
            
            if status in ['PAYMENT_REPROVED_BY_RISK_ANALYSIS', 'PAYMENT_CREDIT_CARD_CAPTURE_REFUSED']:
                #Atualiza o pagamento para saber onde ele esta, mas nao meche no pedido
                pagamento.retornobancario += datetime.datetime.now().strftime("%Y-%m-%d, %H:%M:%S") + '-'+ event_ID + '-' + status +'\n'
                pagamento.status = status
                pagamento.save()
                return 'OK(Credito Negado - tentar novamente)'        
            
            if status in ['PAYMENT_REFUNDED', 'PAYMENT_REFUND_IN_PROGRESS', 'PAYMENT_CHARGEBACK_REQUESTED', 'PAYMENT_CHARGEBACK_DISPUTE', 'PAYMENT_AWAITING_CHARGEBACK_REVERSAL']:
                #Nesse caso e marcar o pagamento para esse status e colocar no pedido  o status de bichado, pq tao brigando
                pagamento.retornobancario += datetime.datetime.now().strftime("%Y-%m-%d, %H:%M:%S") + '-'+ event_ID + '-' + status + '\n'
                pagamento.status = status
                pedido = pagamento.pedido
                pedido.status = 'Disputa'
                pedido.save()
                pagamento.save()
                return 'OK(Disputa de cartao)'
        else:
            return 'error(Retorno Duplicado)'


        #Outros casos que nao tem nada para fazer            
        return 'OK(No Action)'